####################################################################################################
# pimms/calculation.py
# Decorator and class definition for functional calculations.
# By Noah C. Benson

import copy, types, os, sys, re, warnings, pickle, pint, six
import pyrsistent as ps, numpy as np, collections as colls
from functools import reduce
from .util  import (merge, is_pmap, is_str, is_map, is_lazy_map, is_vector, is_quantity, quant, mag,
                    units, qhash, save, load, is_nparray, getargspec_py27like, cache_filename,
                    cache_lmap, qhashform)
from .table import (itable, is_itable)

if six.PY2: tuple_type = types.TupleType
else:       tuple_type = tuple

####################################################################################################
# The Calc, Plan, and IMap classes
class Calc(object):
    '''
    The Calc class encapsulates data regarding the calculation of a single set of data from a
    separate set of input data. The input parameters are referred to as afferent values and the
    output variables are referred to as efferent values.
    '''
    # This is a helper function used in __init__
    @staticmethod
    def _parse_doc(f, affs, effs):
        # pull out the documentation, make it into a map of parameter data for both afferent and
        # efferent values
        res = {'afferent': {}, 'efferent': {}}
        try:
            s = f.__doc__
            assert(isinstance(s, six.string_types))
        except Exception:
            return res
        lines = s.split('\n')
        marks = [re.search('^\s*@\s*([a-zA-Z_]\w*)\s+(.*)$', ln) for ln in lines]
        marks = [None if m is None else (m.group(1), m.group(2)) for m in marks]
        ltype = ['@' if m is not None else 'x' if l == '' else 'c' for (l,m) in zip(lines, marks)]
        dat = [(t,m,l) for (t,m,l) in zip(ltype, marks, lines)]
        while len(dat) > 0:
            pdoc0 = next((i for (i,r) in enumerate(dat) if r[0] == '@'), None)
            if pdoc0 == None: return res
            # okay, there's at least one comment
            pdoc1 = next((i+pdoc0+1 for (i,r) in enumerate(dat[(pdoc0+1):]) if r[0] != 'c'),
                         len(dat))
            # pop the lines
            lns = dat[pdoc0:pdoc1]
            dat = dat[pdoc1:]
            pname = lns[0][1][0]
            txt = pname + ': ' + lns[0][1][1]
            for (_,_,l) in lns[1:]: txt += '\n' + l
            # add this to the appropriate document
            if pname in affs:   (dest, dnm) = (affs, 'afferent')
            elif pname in effs: (dest, dnm) = (effs, 'efferent')
            else:
                warnings.warn('document for unrecognized value %s' % pname)
                continue
            if pname in res[dnm]: res[dnm][pname] += '\n\n' + txt
            else:                 res[dnm][pname] = txt
        # That's it; we can return what's been collected
        return res
    def __init__(self, affs, f, effs, dflts, lazy=True, meta_data={}, cache=False, memoize=True):
        ff_set = set(affs)
        ff_set = ff_set.intersection(effs)
        if ff_set: raise ValueError('calc functions may not overwrite their parameters')
        object.__setattr__(self, 'afferents', affs)
        object.__setattr__(self, 'efferents', effs)
        object.__setattr__(self, 'function',  f)
        object.__setattr__(self, 'defaults',  dflts)
        object.__setattr__(self, 'lazy',      lazy)
        object.__setattr__(self, 'meta_data', ps.pmap(meta_data))
        object.__setattr__(self, 'cache',     cache)
        object.__setattr__(self, 'memoize',   memoize)
        object.__setattr__(self, 'name',      f.__module__ + '.' + f.__name__)
        pdoc = Calc._parse_doc(f, affs, effs)
        pdoc = {k:ps.pmap(v) for (k,v) in six.iteritems(pdoc)}
        object.__setattr__(self, 'afferent_docs', pdoc['afferent'])
        object.__setattr__(self, 'efferent_docs', pdoc['efferent'])
    def __call__(self, *args, **kwargs):
        opts = merge(self.defaults, args, kwargs)
        args = []
        for name in self.afferents:
            if name not in opts:
                raise ValueError('required parameter %s not given to %s' % (name, self.name))
            args.append(opts[name])
        result = self.function(*args)
        if is_map(result):
            if len(result) != len(self.efferents) or not all(e in result for e in self.efferents):
                raise ValueError('keys returned by %s did not match calc declaration' % self.name)
        elif isinstance(result, tuple_type):
            result = {k:v for (k,v) in zip(self.efferents, result)}
        elif len(self.efferents) == 1:
            result = {self.efferents[0]: result}
        elif not self.efferents and not result:
            result = {}
        else:
            raise ValueError(
                'Illegal return value from function call (%s): did not match efferents' % self.name)
        return result
    def __setattr__(self, k, v):
        raise TypeError('Calc objects are immutable')
    def __delattr__(self, k):
        raise TypeError('Calc objects are immutable')
    def set_meta(self, meta_data):
        '''
        node.set_meta(meta) yields a calculation node identical to the given node except that its
        meta_data attribute has been set to the given dictionary meta. If meta is not persistent,
        it is cast to a persistent dictionary first.
        '''
        if not (isinstance(meta_data, ps.PMap) or isinstance(meta_data, IMap)):
            meta_data = ps.pmap(meta_data)
        new_cnode = copy.copy(self)
        object.__setattr__(new_cnode, 'meta_data', meta_data)
        return new_cnode
    def set_defaults(self, *args, **kwargs):
        '''
        node.set_defaults(a=b...) yields a new calculation node identical to the given node except
        with the default values matching the given key-value pairs. Arguments are collapsed left-to
        right with later arguments overwriting earlier arguments.
        '''
        args = merge(self.defaults, args, kwargs)
        new_cnode = copy.copy(self)
        object.__setattr__(new_cnode, 'defaults', ps.pmap(args))
        return new_cnode
    def discard_defaults(self, *args):
        '''
        node.discard_defaults(a, b...) yields a new calculation node identical to the given node
        except that the default values for the given afferent parameters named by the arguments a,
        b, etc. have been removed. In the new node that is returned, these parameters will be
        required.
        '''
        rms = set(arg for aa in args for arg in ([aa] if isinstance(aa, six.string_types) else aa))
        new_defaults = ps.pmap({k:v for (k,v) in six.iteritems(args) if k not in rms})
        new_cnode = copy.copy(self)
        object.__setattr__(new_cnode, 'defaults', new_defaults)
        return new_cnode
    def remove_defaults(self, *args):
        '''
        node.remove_defaults(a, b...) is identical to node.discard_defaults(a, b...) except that
        it raises a KeyError if any of the given arguments are not already defaults.
        '''
        rms = set(arg for aa in args for arg in ([aa] if isinstance(aa, six.string_types) else aa))
        for arg in rms:
            if arg not in self.defaults:
                raise KeyError('{0}'.format(arg))
        return self.discard_defaults(*args)
    def tr(self, *args, **kwargs):
        '''
        calc_fn.tr(...) yields a copy of calc_fn in which the afferent and efferent values of the
        function have been translated. The translation is found from merging the list of 0 or more
        dictionary arguments given left-to-right followed by the keyword arguments. 
        '''
        d = merge(args, kwargs)
        # make a copy
        translation = copy.copy(self)
        object.__setattr__(translation, 'afferents',
                           tuple(d[af] if af in d else af for af in self.afferents))
        object.__setattr__(translation, 'efferents',
                           tuple(d[ef] if ef in d else ef for ef in self.efferents))
        object.__setattr__(translation, 'afferent_docs',
                           ps.pmap({kk:ps.pmap({(d[k] if k in d else k):v
                                                for (k,v) in six.iteritems(vv)})
                                    for (kk,vv) in six.iteritems(self.afferent_docs)}))
        object.__setattr__(translation, 'efferent_docs',
                           ps.pmap({kk:ps.pmap({(d[k] if k in d else k):v
                                                for (k,v) in six.iteritems(vv)})
                                    for (kk,vv) in six.iteritems(self.efferent_docs)}))
        fn = self.function
        def _tr_fn_wrapper(*args, **kwargs):
            res = fn(*args, **kwargs)
            if isinstance(res, colls.Mapping):
                return {(d[k] if k in d else k):v for (k,v) in six.iteritems(res)}
            else:
                return res
        object.__setattr__(translation, 'function', _tr_fn_wrapper)
        return translation
class Plan(object):
    '''
    The Plan class encapsulates individual functions that require parameters as inputs and
    produce outputs in the form of named values. Plan objects can be called as functions with
    a dictionary and/or a list of keyword parameters specifying their parameters; they always return
    a dictionary of the values they calculate, even if they calculate only a single value.

    The Plan class should not be overloaded and should be instantiated using the @calculates
    decorator only; it should not be overloaded directly.
    '''
    # Restricted things:
    def __setattr__(self, k, v):
        raise TypeError('Plan objects are immutable')
    def __detattr__(self, k):
        raise TypeError('Plan objects are immutable')
    # Constructor
    def __init__(self, *args, **kwargs):
        '''
        Plan(nodes...) yields a new Plan object made out of the given keyword list of Calc
        objects, which may be preceded by any number of dictionaries (or plans) whose values are
        calc objects. For each element of the nodes list, if the object is not a Calc, an attempt is
        made to coerce the object to a Calc; if this does not work, an error is raised. Arguments
        are merged left-to-right.
        '''
        nodes = ps.pmap({name: (node if isinstance(node, Calc) else calc(node))
                         for (name,node) in six.iteritems(merge(args, kwargs))})
        if not all(isinstance(n, Calc) for n in six.itervalues(nodes)):
            raise ValueError('All arguments given to Plan must be @calc functions')
        object.__setattr__(self, 'nodes', nodes)
        # let's get the default-value arguments and make sure they all match!
        defaults = {}
        eq = False
        for node in six.itervalues(nodes):
            for (k,v) in six.iteritems(node.defaults):
                if k in defaults:
                    try: eq = np.array_equal(defaults[k], v)
                    except Exception: eq = False
                    if not eq:
                        raise ValueError(
                            'conflicting default values found for \'%s\': %s and %s' % (
                                k, v, defaults[k]))
                else:
                    defaults[k] = v
        # no error yet, so we seem to have agreeing default values at least; note them
        defaults = ps.pmap(defaults)
        object.__setattr__(self, 'defaults', defaults)
        # okay, make the dependency graph
        deps = {}
        affs = set()
        effs = {}
        for node in six.itervalues(self.nodes):
            affs = affs.union(node.afferents)
            effs.update({eff:node for eff in node.efferents})
            for eff in node.efferents:
                if eff in deps:
                    deps[eff].union(node.afferents)
                else:
                    deps[eff] = set(node.afferents)
        for aff in affs:
            if aff not in deps:
                deps[aff] = set([])
        # okay, now find its transitive closure...
        deps0 = ps.pmap({k:tuple(v) for (k,v) in six.iteritems(deps)})
        changed = True
        while changed:
            changed = False
            for (k,v) in six.iteritems(deps):
                if k in v: raise ValueError('self-loop detected in dependency graph (%s)' % k)
                new_set = v.union([depdep for dep in v for depdep in deps[dep] if depdep not in v])
                if new_set != v:
                    changed = True
                    deps[k] = new_set
        deps = ps.pmap({k:tuple(v) for (k,v) in six.iteritems(deps)})
        # alright, deps is now the transitive closure; those afferents that have no dependencies are
        # the calculation's afferent parameters and the efferents are the output values
        affs = tuple(aff for aff in affs if len(deps[aff]) == 0)
        # last, we need to know the afferent dependencies
        aff_deps = ps.pmap(
            {nd.name: tuple(set([aff
                                 for nd_aff in nd.afferents
                                 for aff in ([nd_aff] if nd_aff in affs else deps[nd_aff])
                                 if aff in affs]))
             for nd in nodes.values()})
        object.__setattr__(self, 'afferents', affs)
        object.__setattr__(self, 'efferents', ps.pmap(effs))
        object.__setattr__(self, 'dependencies', deps)
        object.__setattr__(self, 'dependency_edges', deps0)
        object.__setattr__(self, 'afferent_dependencies', aff_deps)
        # we also want to reverse the dependencies so that when an afferent value is edited, we can
        # invalidate the relevant efferent values
        dpts = {}
        for (k,v) in six.iteritems(deps):
            for u in v:
                if u in dpts: dpts[u].add(k)
                else:         dpts[u] = set([k])
        dpts = ps.pmap({k:tuple(v) for (k,v) in six.iteritems(dpts)})
        object.__setattr__(self, 'dependants', dpts)
        # finally, we need to make sure that we know when to call the non-lazy functions; this we
        # do by walking through nodes and picking out all the fully afferent dependencies of each
        reqs = {}
        zero_reqs = []
        for node in six.itervalues(nodes):
            if node.lazy:            continue               # we only care about non-lazy nodes...
            elif not node.afferents: zero_reqs.append(node) # this node depends on nothing...
            else:                                           # otherwise, find the afferent deps...
                # we make a set of all upstream afferent values then intersect it with the
                # calculation's afferent values 
                node_affs = set(node.afferents)
                for aff in node.afferents:
                    node_affs = node_affs.union(deps[aff])
                node_affs &= set(self.afferents)
                # okay, add this node to the relevant afferent requirements lists
                for aff in node_affs:
                    if aff in reqs: reqs[aff].append(node)
                    else:           reqs[aff] = [node]
        for aff in self.afferents:
            if aff not in reqs:
                reqs[aff] = []
        reqs = ps.pmap({k:tuple(v) for (k,v) in six.iteritems(reqs)})
        zero_reqs = tuple(zero_reqs)
        object.__setattr__(self, 'proactive_dependants', reqs)
        object.__setattr__(self, 'initializers', zero_reqs)
        # Okay, now that we have everything else organized, lets build the documentation for the
        # afferent and efferent data
        # Afferent docs first:
        adocs = {}
        for aff in affs:
            txt = aff
            if aff in defaults: txt += ' (default: %s)' % defaults[aff]
            for (nnm, node) in six.iteritems(nodes):
                if aff not in node.afferent_docs: continue
                txt += ('\n\n(%s) ' % nnm) + node.afferent_docs[aff]
            adocs[aff] = txt
        # Then efferents:
        edocs = {}
        for eff in effs:
            txt = eff
            for (nnm, node) in six.iteritems(nodes):
                if eff not in node.efferent_docs: continue
                txt += ('\n\n(%s) ' % nnm) + node.efferent_docs[eff]
            edocs[eff] = txt
        object.__setattr__(self, 'afferent_docs', ps.pmap(adocs))
        object.__setattr__(self, 'efferent_docs', ps.pmap(edocs))
        # Make a value for memoized data:
        object.__setattr__(self, '_memoized_data', {})
        # That's it; we should be constructed now!
    def __call__(self, *args, **kwargs):
        '''
        cplan(args...) runs the given calculation plan cplan on the given args and returns a
        dictionary of the yielded values.
        '''
        return IMap(self, merge(self.defaults, args, kwargs))
    def _check(self, calc_dict, changes=None):
        '''
        cplan._check(calc_dict) should be called only by the calc_dict object itself.
        The check method makes sure that all of the proactive methods on the calc_dict are run; if
        the optional keyword argument changes is given, then checks are only run for the list of
        changes given.
        '''
        if changes is None:
            changes = self.afferents
            # run the zero-reqs
            for req in self.initializers:
                calc_dict._run_node(req)
        else:
            # invalidate these data if needed
            dpts = set(dpt for aff in changes for dpt in self.dependants[aff])
            for dpt in dpts: del calc_dict[dpt]
        # now, make sure that proactive dependants get run
        proactives = set(node for change in changes for node in self.proactive_dependants[change])
        for node in proactives:
            calc_dict._run_node(node)
        # That's it!
    def set(self, **kwargs):
        '''
        cplan.set(a=b...) yields a new caclulation plan identical to cplan except such that the
        calculation pieces specified by the arguments have been replaced with the given
        calculations instead.
        '''
        return Plan(reduce(lambda m,u: m.set(u[0],u[1]), six.iteritems(kwargs), self.nodes))
    def discard(self, *args):
        '''
        cplan.discard(...) yields a new calculation plan identical to cplan except without any of
        the calculation steps listed in the arguments.
        '''
        return Plan(reduce(lambda m,k: m.discard(k), args, self.nodes))
    def remove(self, *args):
        '''
        cplan.remove(...) yields a new calculation plan identical to cplan except without any of
        the calculation steps listed in the arguments. An exception is raised if any keys are not
        found in the calc-plan.
        '''
        return Plan(reduce(lambda m,k: m.remove(k), args, self.nodes))
    def set_defaults(self, *args, **kwargs):
        '''
        cplan.set_defaults(a=b...) yields a new caclulation plan identical to cplan except such
        that the calculation default values specified by the arguments have been replaced with the
        given values instead. E.g., cplan.set_defaults(a=1) would return a new plan with a default
        value for a=1.
        '''
        d = merge(args, kwargs)
        # make a copy of this object's nodes with translations...
        nodes = ps.pmap({k:v.set_defaults(d) for (k,v) in six.iteritems(self.nodes)})
        # make a new plan with that!
        return Plan(nodes)
    def discard_defaults(self, *args):
        '''
        cplan.discard_defaults(a, b...) yields a new caclulation plan identical to cplan except 
        without default values for any of the given parameter names.
        '''
        # make a copy of this object's nodes with translations...
        nodes = ps.pmap({k:v.discard_defaults(*args) for (k,v) in six.iteritems(self.nodes)})
        # make a new plan with that!
        return Plan(nodes)        
    def remove_defaults(self, *args):
        '''
        cplan.remove_defaults(a, b...) yields a new caclulation plan identical to cplan except 
        without default values for any of the given parameter names. An exception is raised if any
        default value given is not found in cplan.
        '''
        for arg in args:
            if arg not in self.defaults:
                raise KeyError('{0}'.format(arg))
        return self.discard_defaults(*args)
    def tr(self, *args, **kwargs):
        '''
        p.tr(...) yields a copy of plan p in which the afferent and efferent values of all of the
        calc functions contained in the plan have been translated. The translation is found from
        merging the list of 0 or more dictionary arguments given left-to-right followed by the
        keyword arguments.
        '''
        d = merge(args, kwargs)
        # make a copy of this object's nodes with translations...
        nodes = ps.pmap({k:v.tr(d) for (k,v) in six.iteritems(self.nodes)})
        # make a new plan with that!
        return Plan(nodes)
    def forget(self, node=None, cache_directory=None):
        '''
        plan.forget() clears the in-memory memoized cache for the plan and returns the cache dict
          prior to clearing.
        '''
        #plan.forget(node) clears only the cache for the given node (which may be a node name) and
        #  all nodes that depend on it; plan.forget(None) is equivalent to plan.forget()
        #plan.forget(node, cache_directory) additionally deletes all directories in the given cache
        #  directory that correspond to the nodes downstream of node. The node may be None to delete
        #  all relevant cache directories.
        m = self._memoized_data
        self._memoized_data = {}
        return m
        
class IMap(colls.Mapping):
    '''
    The IMap class instantiates a lazy immutable mapping from both parameters and calculated
    value names to their values. IMap objects should only be created in two ways:
     (1) by passing a calculation plan a complete set of parameters, or,
     (2) by calling the using method on an existing IMap to update the parameters.
    '''
    def __init__(self, plan, afferents):
        object.__setattr__(self, 'plan', plan)
        object.__setattr__(self, 'afferents', ps.pmap(afferents))
        object.__setattr__(self, 'efferents', ps.m())
        # We need to run the checks from the calculation
        plan._check(self)
        # otherwise, we're set!
    # IMap does not allow certain things:
    def __setattr__(self, key, val):
        raise TypeError('IMap objects are immutable')
    def __delattr__(self, key):
        raise TypeError('IMap objects are immutable')
    # The Mapping functions that must be overloaded:
    def __len__(self):
        return len(self.afferents) + len(self.plan.efferents)
    def __iter__(self):
        for k in six.iterkeys(self.afferents):
            yield k
        for k in six.iterkeys(self.plan.efferents):
            yield k
    def __getitem__(self, k):
        if k in self.afferents:
            return self.afferents[k]
        elif k not in self.plan.efferents:
            raise ValueError('Key \'%s\' not found in calc-dictionary' % k)
        else:
            # using get instead of 'if k in efferents: return ...' avoids the race-condition that
            # would otherwise cause problems when someone deletes an item from the cache using
            # delitem below
            effval = self.efferents.get(k, self)
            if effval is self:
                self._run_node(self.plan.efferents[k])
                return self.efferents[k]
            else:
                return effval
    get = colls.Mapping.get
    # We want the representation to look something like a dictionary
    def __repr__(self):
        affstr = ', '.join([repr(k) + ': ' + repr(v) for (k,v) in six.iteritems(self.afferents)])
        effstr = ', '.join([repr(k) + ': <lazy>' for k in six.iterkeys(self.plan.efferents)])
        if len(self.afferents) == 0:
            return 'imap({' + effstr + '})'
        elif len(self.plan.efferents) == 0:
            return 'imap({' + affstr + '})'
        else:
            return 'imap({' + affstr + ', ' + effstr + '})'
    # There are a few other methods we want to be careful of also:
    def __contains__(self, k):
        return (k in self.afferents) or (k in self.plan.efferents)
    def __delitem__(self, k):
        '''
        IMap's __delitem__ method allows one to clear the cached value of an efferent.
        '''
        if k in self.afferents:
            raise TypeError('Cannot delete a parameter (%s) from a IMap' % k)
        elif k in self.efferents:
            object.__setattr__(self, 'efferents', self.efferents.discard(k))
        elif k not in self.plan.efferents:
            raise TypeError('IMap object has no item named \'%s\'' % k)
        # else we don't worry about it; not yet calculated.
    #indent = ' -' #dbg
    def _run_node(self, node):
        '''
        calc_dict._run_node(node) calculates the results of the given calculation node in the
        calc_dict's calculation plan and caches the results in the calc_dict. This should only
        be called by calc_dict itself internally.
        '''
        #print IMap.indent, ('Node: %s' % node.name) #dbg
        #IMap.indent = '  ' + IMap.indent #dbg
        #
        # We need to pause here and handle caching, if needed.
        (res, h, found) = (None, None, False)
        if ('memoize' in self.afferents and self.afferents['memoize']) and node.memoize:
            memdat = self.plan._memoized_data
            try:
                h = {k:self.afferents[k] for k in self.plan.afferent_dependencies[node.name]}
                h = qhashform((node.name, ps.pmap(h)))
                if h in memdat:
                    # memoization success; set h to None so we don't re-memoize the value
                    (res, h, found) = (memdat[h], None, True)
                elif node.cache:
                    cpath = self.afferents.get('cache_directory', None)
                    if cpath is not None:
                        ureg = self.afferents.get('unit_registry', 'pimms')
                        try:
                            cpath = cache_filename(cpath, h)
                            if os.path.isfile(cpath):
                                # set cpath to None to signal we don't need to cache the value out
                                (res, cpath, found) = (load(cpath, ureg), None, True)
                        except Exception: cpath = None
                else: cpath = None
            except Exception: raise#b(h, found) = (None, False)
        # ensure we have a result
        if not found: res = node(self)
        # process the result:
        effs = reduce(lambda m,v: m.set(v[0],v[1]), six.iteritems(res), self.efferents)
        object.__setattr__(self, 'efferents', effs)
        # Handle the caching if needed:
        if h is not None:
            memdat[h] = res
            #if cpath is None: print IMap.indent, 'hashed' #dbg
            if cpath is not None:
                try: save(cpath, res)
                except Exception: pass
        #elif node.memoize: print IMap.indent, 'simple-calc' #dbg
        #IMap.indent = IMap.indent[2:] #dbg
    # Also, since we're a persistent object, we should follow the conventions for copying
    def set(self, *args, **kwargs):
        '''
        d.set(...) yields a copy of the IMap object d; the ... may be replaced with either
        nothing (in which case d is returned) or a list of 0 or more dictionaries followed by a lsit
        of zero or more keyword arguments. These dictionaries and keywords arguments are merged
        left-to-right; the result may contain only afferent parameters of d and replaces the values
        of d in the newly returned calc dictionary.
        '''
        args = merge(args, kwargs)
        if len(args) == 0: return self
        affs = self.afferents
        pln = self.plan
        # make sure these are all valid parameters
        if any(k not in affs for k in six.iterkeys(args)):
            raise TypeError(
                'The given key \'%s\' is not an afferent parameter of IMap object')
        # okay, we can make the change...
        new_calc_dict = copy.copy(self)
        new_affs = reduce(lambda m,v: m.set(v[0],v[1]), six.iteritems(args), affs)
        object.__setattr__(new_calc_dict, 'afferents', new_affs)
        # we need to run checks and delete any cache that has been invalidated.
        # The calculation's check method does this; it raises an exception if there is an error
        pln._check(new_calc_dict, changes=args)
        return new_calc_dict
    def tr(self, *args, **kwargs):
        '''
        imap.tr(...) yields a copy of the immutable map imap in which both the plan and the keys of
        the new map have been translated according to the translation given in the arguments list.
        The translation is found from merging the list of 0 or more dictionary arguments given
        left-to-right followed by the keyword arguments.
        '''
        d = merge(args, kwargs)
        # make a copy of the plan first:
        new_plan = self.plan.tr(d)
        # now use that plan to re-initialize ourself
        return new_plan(**{(d[k] if k in d else k):v for (k,v) in six.iteritems(self.afferents)})

####################################################################################################
# Identification functions for these types
def is_calc(arg):
    '''
    is_calc(x) yields True if x is a function that was decorated with an @calc directive and False
    otherwise.
    '''
    return isinstance(arg, Calc)
def is_plan(arg):
    '''
    is_plan(x) yields True if x is a calculation plan made with the plan function and False
    otherwise.
    '''
    return isinstance(arg, Plan)
def is_imap(arg):
    '''
    is_imap(x) yields True if x is an IMap object and False otherwise.
    '''
    return isinstance(arg, IMap)
    
####################################################################################################
# Creation function for Calc, Plan, and IMap objects
def calc(*args, **kwargs):
    '''
    @calc is a decorator that indicates that the function that follows is a calculation component;
      calculation components can be used to form Plan objects (see Plan). In this
      case, the return value is given the same name as the function.
    @calc(names...) accepts a string or list/tuple of strings that name the output values of the
      calc function. In this case, the function must return either a tuple of thes values in the
      order or a dictionary in which the keys are the same as the given names. In this case, the
      optional value lazy=False may be passed after the names to indicate that the calculation
      should be run immediately when parameters are set/changed in a calculation rather than lazily
      when requested.
    @calc(None) is a special instance in which the lazy argument is ignored, no efferent values are
      expected, and the calculation is always run when the afferent parameters are updated.
    '''
    # parse out the only accepted keywords args:
    opt_names = ['lazy', 'meta', 'cache', 'memoize']
    opt_dflts = [True,   {},     False,   True]
    (lazy, meta, cache, mem) = [kwargs.get(nm,df) for (nm,df) in zip(opt_names, opt_dflts)]
    if len(kwargs) != len([k for k in opt_names if k in kwargs]):
        raise ValueError('calc accepts only the options lazy, meta, and cache')
    if len(args) == 1 and not isinstance(args[0], six.string_types):
        if isinstance(args[0], types.FunctionType):
            f = args[0]
            effs = (f.__name__,)
            (affs, varargs, kwargs, dflts) = getargspec_py27like(f)
            if varargs or kwargs: raise ValueError('@calc functions may not accept variadic args')
            affs = tuple(affs)
            dflts = ps.pmap({} if dflts is None else
                            {k:v for (k,v) in zip(affs[-len(dflts):], dflts)})
            kwargs = {} if not kwargs else kwargs
            return Calc(affs, f, effs, dflts, **kwargs)
        elif args[0] is None:
            effs = ()
            def _calc_req(f):
                (affs, varargs, kwargs, dflts) = getargspec_py27like(f)
                if varargs or kwargs: raise ValueError('@calc functions only accept simple params')
                affs = tuple(affs)
                dflts = ps.pmap({} if dflts is None else
                                {k:v for (k,v) in zip(affs[-len(dflts):], dflts)})
                return Calc(affs, f, effs, dflts,
                            lazy=False, meta_data=meta, cache=cache, memoize=mem)
            return _calc_req
        else:
            raise ValueError('calc only accepts strings, None, or no argument')
    elif len(args) < 1:
        raise ValueError('calc should be used as a function decorator')
    elif not all(isinstance(arg, six.string_types) for arg in args):
        raise ValueError('@calc(...) requires that all arguments be keyword strings')
    else:
        effs = tuple(args)
        kwargs0 = kwargs
        def _calc(f):
            (affs, varargs, kwargs, dflts) = getargspec_py27like(f)
            if varargs or kwargs:
                raise ValueError('@calc functions may only accept simple parameters')
            affs = tuple(affs)
            dflts = ps.pmap({} if dflts is None else
                            {k:v for (k,v) in zip(affs[-len(dflts):], dflts)})
            if kwargs is None: kwargs = {}
            return Calc(affs, f, effs, dflts, **kwargs0)
        return _calc
def plan(*args, **kwargs):
    '''
    plan(name1=calcs1, name2=calc2...) yields a new calculation plan (object of type
      Plan) that is itself a constructor for the calculation dictionary that is implied by
      the given calc functionss given. The names that are given are used as identifiers for
      updating the calc plan (using the without and using methods).
    plan(arg1, arg2..., name1=calc1, name2=calc2...) additionally initializes the dictionary
      of calculations and names with the calculation plans or dictionaries given as arguments. These
      are collapsed left-to-right.
    plan(imap) yields the plan object for the given IMap imap.
    '''
    if len(args) == 1 and len(kwargs) == 0 and is_imap(args[0]):
        return args[0].plan
    adict = merge(tuple(arg.nodes if isinstance(arg, Plan) else arg for arg in args),
                  kwargs)
    return Plan(**adict)
def planfn(*args, **kwargs):
    '''
    planfn(val1=fn1, val2=fn2...) uses the pimms plan mechanism to yield a function f that produces
      an immutable imap when called with the correct paramters. Unlike in plan(), planfn() does not
      care about the names of the calculation units; instead the val1, val2, etc. are names of the
      efferent values while the given fn objects are the functions that calculate them (they are
      automatically converted into calc objects, and names are automatically generated for the items
      required).
    planfn(res, val1=fn1, val2=fn2...) yields a function f that is equivalent to the function
      lambda *args,**kw: g(*args,**kw)[res] where g = planfn(val1=fn1, val2=fn2...).
    planfn((p1, p2...), val1=fn1, val2=fn2...) yields a function f that is equivalent to the
      function lambda *a: g({p1:a[0], p2:a[1]...}) where g = planfn(val1=fn1, val2=fn2...).
    planfn((p1, p2...), res, val1=fn1, val2=fn2...) applies both of the above translations to the
      resulting function; note that either the first or second argument may be None to specify
      default behavior. Additionally, the first argument may be a string s equivalent to (s,).
    '''
    # first: parse options
    if   len(args) > 2:           raise ValueError('no more than 2 non-keyword arguments allowed')
    elif len(args) == 0:          (params,result) = (None,None)
    elif len(args) == 2:          (params,result) = args
    elif is_str(args[0]):         (params,result) = (None,args[0])
    elif is_vector(args[0], str): (params,result) = (args[0],None)
    else:                         raise ValueError('cannot interpret argument: %s' % args[0])
    # okay, now, we make the calcs and the calc plan:
    calcs = {}
    for (k,v) in six.iteritems(kwargs):
        # if v is already a calc object, we can only use this if it has exactly one efferent
        if is_calc(v):
            if len(v.efferents) != 1:
                raise ValueError('calc objects given to planfn must have exactly 1 efferent')
            eff = v.efferents[0]
            if eff != k: v = v.tr({eff:k})
        else: v = calc(k)(v) # otherwise convert it to a calc with the output we want
        # it's possible that the names of the calc objects are not unique, so we want to auto-
        # generate unique names based on their names...
        calcs['calc_' + k] = v
    p0 = plan(calcs)
    # see if we need to fix the input
    if params is None:           p = p0
    elif is_str(params):         p = lambda arg: p0({params:arg})
    elif is_vector(params, str): p = lambda *args: p0({k:v for (k,v) in zip(params,args)})
    else: raise ValueError('params must be a string or list of strings or None')
    # and if we need to fix the output
    if result is None: return p
    def f(*args, **kw): return p(*args, **kw)[result]
    return f
def imap(p, *args, **kwargs):
    '''
    imap(p, args...) yields an immutable map object made from the plan object p and the given
    arguments, which may be any number of mappings followed by any number of keyword arguments,
    all of which are merged left-to-right then interpreted as the parameters of the given plan p.
    '''
    p = p if is_plan(p) else plan(p)
    params = merge(args, kwargs)
    return p(params)

####################################################################################################
# Functions for updating the immutable types.
def calc_tr(calc_fn, *args, **kwargs):
    '''
    calc_tr(calc_fn, ...) yields a copy of calc_fn in which the afferent and efferent values of the
      function have been translated. The translation is found from merging the list of 0 or more
      dictionary arguments given left-to-right followed by the keyword arguments. If the calc_fn
      that is given is not a @calc function explicitly, calc_tr will attempt to coerce it to one.
    '''
    if not is_calc(calc_fn):
        calc_fn = calc(calc_fn)
    return calc_fn.tr(*args, **kwargs)
def plan_tr(p, *args, **kwargs):
    '''
    plan_tr(p, ...) yields a copy of plan p in which the afferent and efferent values of its
      functions have been translated. The translation is found from merging the list of 0 or more
      dictionary arguments given left-to-right followed by the keyword arguments. If the plan that
      is given is not a plan object explicitly, calc_tr will attempt to coerce it to one.
    '''
    if not is_plan(p):
        p = plan(p)
    return p.tr(*args, **kwargs)
def imap_tr(imap, *args, **kwargs):
    '''
    imap_tr(m, ...) yields a copy of the immutable map m in which the keywords have been translated
      according to the given arguments. Arguments may be any number of dictionaries followed by any
      number of keyword arguments, all of which are merged left-to-right then used as the
      translation.
    '''
    if not is_imap(imap):
        raise TypeError('IMap object required of imap_tr')
    return imap.tr(*args, **kwargs)




# Usage example: calculating the area of a triangle
import pimms

# We make a lazy calculation plan to calculate the area of a triangle; first it calculates the base
# and height, then the area.

# First calc unit: calculate the base and the height
@pimms.calc('base', 'height')
def calc_triangle_base(a, b, c):
    '''
    calc_triangle_base computes the base (x-width) of the triangle a-b-c.
    
    Afferent parameters:
     @ a Must be the (x,y) coordinate of point a in triangle a-b-c.
     @ b Must be the (x,y) coordinate of point b in triangle a-b-c.
     @ c Must be the (x,y) coordinate of point c in triangle a-b-c.

    Efferent values:
     @ base Will be the base, or width, of the triangle a-b-c.
     @ height Will be the height of the triangle a-b-c.
    '''
    print 'Calculating base...'
    xs = [a[0], b[0], c[0]]
    xmin = min(xs)
    xmax = max(xs)
    print 'Calculating height...'
    ys = [a[1], b[1], c[1]]
    ymin = min(ys)
    ymax = max(ys)
    return (xmax - xmin, ymax - ymin)

# Second calc unit: calculate the area
@pimms.calc('area')
def calc_triangle_area(base, height):
    '''
    calc_triangle_are computes the area of a triangle with a given base and height.
    
    Efferent values:
     @ area Will be the area of the triangle with the given base and height.
    '''
    print 'Calculating area...'
    return {'area': base * height * 0.5}



#! /usr/bin/env python
####################################################################################################

from setuptools import (setup, Extension)

setup(
    name='pimms',
    version='0.3.10',
    description='Python immutable data structures library',
    keywords='persistent immutable functional',
    author='Noah C. Benson',
    author_email='nben@nyu.edu',
    url='https://github.com/noahbenson/pimms/',
    license='GPLv3',
    packages=['pimms', 'pimms.test'],
    #ext_modules=[Extension('pimms.c_util', ['pimms/c_util.c'])],
    package_data={'': ['LICENSE.txt']},
    include_package_data=True,
    install_requires=['pyrsistent>=0.11',
                      'six>=1.10',
                      'numpy>=1.2',
                      'pint>=0.7'])



####################################################################################################
# pimms/table.py
# Classes for storing immutable data tables.
# By Noah C. Benson

import copy, types, sys, pint, six
import numpy                      as     np
import pyrsistent                 as     ps
from   functools                  import reduce
from   .util                      import (merge, is_pmap, is_map, LazyPMap, lazy_map, is_lazy_map,
                                          is_quantity, is_unit, is_str, is_int, is_vector,
                                          quant, iquant, mag, unit, qhash, units, imm_array,
                                          getargspec_py27like)
from   .immutable                 import (immutable, value, param, require, option)

if sys.version_info[0] == 3: from   collections import abc as colls
else:                        import collections            as colls

def _ndarray_assoc(arr, k, v):
    '_ndarray_assoc(arr, k, v) duplicates arr to a writeable array, sets arr2[k]=v, returns arr2'
    arr = np.array(arr)
    arr[k] = v
    arr.setflags(write=False)
    return arr    

class ITableRow(colls.Mapping):
    '''
    ITableRow is a class that works with the ITable class to quickly and lazily allow access to
    individual rows as if they were individual persistent maps. For all intents and purposes, an
    ITableRow object should be treated as a dict object that cannot be changed.
    Note that ITableRow is not an immutable class, but its members cannot be changed. The class
    is intended as a hidden subclass that is very efficient.
    '''
    def __init__(self, data, colnames, rownum):
        object.__setattr__(self, 'data',         data)
        object.__setattr__(self, 'column_names', colnames)
        object.__setattr__(self, 'row_number',   rownum)
    def keys(self):
        return self.column_names
    def __setattr__(self, k, v):
        raise RuntimeError('ITableRow object is immutable')
    def __getitem__(self, key):
        return self.data[key][self.row_number]
    def __setitem__(self, key, value):
        raise RuntimeError('Cannot set row of immutable table')
    def __delitem__(self, key):
        raise RuntimeError('Cannot set row of immutable table')
    def __iter__(self):
        dat = self.data
        n = self.row_number
        for col in self.column_names:
            yield col
    def __len__(self):
        return len(self.column_names)
    def asdict(self):
        return {k:self.data[k][self.row_number] for k in self.__iter__()}
    def aspmap(self):
        return ps.pmap(self.asdict())
    def __repr__(self):
        return repr(self.asdict())
    def __hash__(self):
        return hash(self.aspmap())

@immutable
class ITable(colls.Mapping):
    '''
    The ITable class is a simple immutable datatable.
    '''
    def __init__(self, data, n=None):
        self.data = data
        self._row_count = n
    def __hash__(self):
        # we want to make sure arrays are immutable
        return qhash(self.data)
    def __getstate__(self):
        d = self.__dict__.copy()
        d['data'] = {k:(mag(v), unit(v)) if is_quantity(v) else (v, None)
                     for (k,v) in six.iteritems(self.data)}
        return d
    def __setstate__(self, d):
        dat = d['data']
        object.__setattr__(self, 'data',
                           ps.pmap({k:(imm_array(u) if v is None else iquant(u, v))
                                    for (k,(u,v)) in six.iteritems(dat)}))
        object.__setattr__(self, '_row_count', None)
    @staticmethod
    def _filter_col(vec):
        '_filter_col(vec) yields a read-only numpy array version of the given column vector'
        if isinstance(vec, types.FunctionType) and getargspec_py27like(vec)[0] == []:
            return lambda:ITable._filter_col(vec())
        elif is_quantity(vec):
            m = mag(vec)
            mm = ITable._filter_col(m)
            return vec if m is mm else quant(mm, unit(vec))
        else:
            return imm_array(vec)
    @param
    def data(d):
        '''
        itbl.data is an immutable map of the given itable in which property names are associated
        with their data vectors.
        '''
        # we want to check these values and clean them up as we go, but if this is a lazy map, we
        # want to do that lazily...
        if is_map(d):
            if not is_lazy_map(d): d = lazy_map(d)
            def _make_lambda(k): return (lambda:ITable._filter_col(d[k]))
            return lazy_map(
                {k:_make_lambda(k) if d.is_lazy(k) else ITable._filter_col(d[k])
                 for k in six.iterkeys(d)})
        else:
            raise ValueError('Unable to interpret data argument; must be a mapping')
    @param
    def _row_count(n):
        '''
        itbl._row_count is the row count, as provided by internal methods when the row count can be
        known ahead of time. It should not geberally be used; use itbl.row_count instead.
        '''
        return n
    @require
    def validate_data(data):
        '''
        ITable data is required to be a PMap with keys that are strings.
        '''
        if not isinstance(data, ps.PMap):
            raise ValueError('data is required to be a persistent map')
        if not all(isinstance(k, six.string_types) for k in six.iterkeys(data)):
            raise ValueError('data keys must be strings')
        return True
    @require
    def validate_row_count(_row_count):
        '''
        ITable _row_count must be a non-negative integer or None.
        '''
        if _row_count is None: return True
        else: return is_int(_row_count) and _row_count >= 0
    @value
    def column_names(data):
        '''
        itbl.column_names is a tuple of the names of the columns of the data table.
        '''
        return tuple(six.iterkeys(data))
    @value
    def row_count(data, _row_count):
        '''
        itbl.row_count is the number of rows in the given datatable itbl.
        '''
        if len(data) == 0:
            return 0
        elif _row_count:
            return _row_count
        elif is_lazy_map(data):
            # if data is a lazy map, we look first for a column that isn't lazy:
            k = next(data.iternormal(), None)
            k = k if k else next(data.itermemoized(), None)
            k = k if k else next(data.iterkeys())
            return len(data[k])
        else:
            return len(next(six.itervalues(data), []))
    @value
    def columns(data, row_count):
        '''
        itbl.columns is a tuple of the columns in the given datatable itbl. Anything that depends on
        columns includes a de-facto check that all columns are the same length.
        '''
        cols = tuple(v for v in six.itervalues(data))
        if not all(len(c) == row_count for c in cols):
            raise ValueError('itable columns do not all have identical lengths!')
        return cols
    @value
    def rows(data, row_count, column_names):
        '''
        itbl.rows is a tuple of all the persistent maps that makeup the rows of the data table.
        '''
        return tuple([ITableRow(data, column_names, i) for i in range(row_count)])
    @value
    def dataframe(data):
        '''
        itbl.dataframe is a pandas dataframe object that is equivalent to the given itable. Note
          you must have pandas installed for this to work; an exception will be raised when this
          value is requested if you do not.
        '''
        import pandas
        return pandas.DataFrame.from_dict(dict(data))
    # Methods
    def set(self, k, v):
        '''
        itbl.set(name, val) yields a new itable object identical to the given itbl except that it
          includes the vector val under the given column name.
        itbl.set(row, map) updates just the given row to have the properties in the given map; if
          this results in a new column being added, it will have the value None for all other rows.
        itbl.set(rows, m) allows a sequence of rows to be set by passing rows as either a list or
          slice; m may also be a single map or a sequence of maps whose size matches that of rows.
          Alternately, m may be an itable whose row-size matches that of rows; in this case new
          column names may again be added.
        '''
        dat = self.data
        if isinstance(k, six.string_types):
            if isinstance(v, (ITable, colls.Mapping)): v = v[k]
            v = self._filter_col(v)
            new_data = self.data.set(k, v)
            return ITable(new_data, n=len(v))
        elif is_int(k):
            # This is an awful slow way to do things
            def _make_lambda(k):
                return lambda:_ndarray_assoc(dat[k], k, v[k]) if k in v else dat[k]
            new_map = {k:_make_lambda(k) for k in six.iterkeys(dat)}
            nones = np.full((self.row_count,), None)
            for (vk,v) in six.iteritems(v):
                if vk not in new_map:
                    new_map[vk] = _ndarray_assoc(nones, k, v)
            return ITable(lazy_map(new_map), n=self.row_count)
        elif not k:
            return self
        elif isinstance(k[0], six.string_types):
            nones = np.full((self.row_count,), None)
            newdat = self.data
            if isinstance(v, ITable):
                def _make_lambda(k): return (lambda:self._filter_col(v[kk]))
                v = lazy_map({kk:_make_lambda(kk) for kk in k})
            elif not isinstance(v, colls.Mapping):
                v = np.asarray(v)
                if len(v) == self.row_count and v.shape[1] == len(k): v = v.T
                v = {kk:self._filter_col(vv) for (kk,vv) in zip(k,v)}
            for kk in six.iterkeys(v):
                def _make_lambda(k): return (lambda:self._filter_col(v[kk]))
                newdat = newdat.set(kk, _make_lambda(kk) if kk in v else nones)
            return ITable(newdat, n=self.row_count)
        else:
            (keys, vals) = (k,v)
            dat = self.data
            nones = np.full((self.row_count,), None)
            knones = np.full((len(keys),), None)
            if isinstance(vals, (ITable, colls.Mapping)):
                def _make_lambda(k):
                    return lambda:_ndarray_assoc(
                        dat[k] if k in dat else nones,
                        keys,
                        vals[k] if k in vals else knones)
                dat = reduce(
                    lambda m,k: m.set(k, _make_lambda(k)),
                    six.iteritems(vals.data if isinstance(vals, ITable) else vals),
                    dat)
            else:
                def _make_lambda(k): return lambda:np.asarray([v[k] for v in vals])
                cols = lazy_map({k:_make_lambda(k) for k in six.iterkeys(vals[0])})
                def _make_lambda(k):
                    return lambda:_ndarray_assoc(
                        dat[k] if k in dat else nones,
                        keys,
                        cols[k])
                dat = reduce(
                    lambda m,k: m.set(k, _make_lambda(k)),
                    six.iterkeys(vals[0]),
                    dat)
            return ITable(dat, n=self.row_count)
    def discard(self, cols):
        '''
        itbl.discard(arg) discards either the list of rows, given as ingtegers, or the list of
          columns, given as strings.
        '''
        if not cols: return self
        dat = self.data
        vecq = is_vector(cols)
        if is_str(cols) or (vecq and len(cols) > 0 and is_str(cols[0])):
            cols = set(cols if vecq else [cols])
            def _make_lambda(k): return lambda:dat[k]
            return ITable(lazy_map({k:_make_lambda(k) for k in six.iterkeys(dat) if k not in cols}),
                          n=self.row_count)
        elif isinstance(cols, slice) or is_int(cols) or \
             (vecq and len(cols) > 0 and is_int(cols[0])):
            def _make_lambda(k): return lambda:np.delete(dat[k], cols, 0)
            newdat = lazy_map({k:_make_lambda(k) for k in six.iterkeys(dat)})
            return ITable(newdat, n=len(np.delete(np.ones((self.row_count,)), cols, 0)))
        elif vecq and len(cols) == 0: return self
        else: raise ValueError('ITable.discard requires integers or strings')
    def is_lazy(self, k):
        '''
        itable.is_lazy(k) yields True if k is a lazy value in the given itable, as in a lazy map.
        '''
        return self.data.is_lazy(k)
    def is_memoized(self, k):
        '''
        itable.is_memoized(k) yields True if k is a memoized value in the given itable, as in a lazy
          map.
        '''
        return self.data.is_memoized(k)
    def is_normal(self, k):
        '''
        itable.is_normal(k) yields True if k is a normal value in the given itable, as in a lazy
          map.
        '''
        return self.data.is_normal(k)
    def iterkeys(self):
        return self.data.iterkeys()
    def iteritems(self):
        return self.data.iteritems()
    def iterlazy(self):
        '''
        itable.iterlazy() yields an iterator over the lazy keys only (memoized lazy keys are not
          considered lazy).
        '''
        return self.data.iterlazy()
    def itermemoized(self):
        '''
        itable.itermemoized() yields an iterator over the memoized keys only (neihter unmemoized
          lazy keys nor normal keys are considered memoized).
        '''
        return self.data.itermemoized()
    def iternormal(self):
        '''
        itable.iternormal() yields an iterator over the normal unlazy keys only (memoized lazy keys
          are not considered normal).
        '''
        return self.data.iternormal()
    def map(self, f):
        '''
        itbl.map(f) yields the result of mapping the rows of the given datatable itbl over the
          given function f.
        '''
        if isinstance(f, six.string_types) and f in self.data: return self.data[f]
        (args, vargs, kwargs, dflts) = getargspec_py27like(f)
        dflts = dflts if dflts else ()
        dflts = tuple([None for _ in range(len(args) - len(dflts))]) + dflts
        # we have to decide what to map over...
        return map(f, self.rows)
    def where(self, f):
        '''
        itbl.where(f) yields the indices for which itbl.map(f) yields True.
        '''
        return [i for (i,v) in enumerate(self.map(f)) if v]
    def select(self, arg):
        '''
        itbl.select(idcs) yields a sub-table in which only the rows indicated by the given list of
          indices are kept.
        itbl.select(f) keeps all rows for which the function f yields True.
        '''
        if isinstance(arg, types.FunctionType):
            arg = self.where(arg)
        else:
            n = len(arg)
            if n == self.row_count and set(arg) == set([0,1]):
                arg = [i for (i,b) in enumerate(arg) if b]
                n = len(arg)
            dat = self.data
            def _make_lambda(k): return lambda:dat[k][arg]
            return ITable(
                lazy_map({k:_make_lambda(k) for k in six.iterkeys(dat)}),
                n=n)
    def merge(self, *args, **kwargs):
        '''
        itbl.merge(...) yields a copy of the ITable object itbl that has been merged left-to-right
          with the given arguments.
        '''
        return itable(self.data, *args, **kwargs).persist()
    def __getitem__(self, rows, cols=Ellipsis):
        '''
        itbl[row_number] yields the map associated with the given row in the ITable object itbl; the
          row_number may alternately be a slice.
        itbl[[r1, r2...]] yields a duplicate itable containing only the given rows of itbl.
        itbl[column_name] yields the numpy array associated with the given column name.
        itbl[[c1, c2...]] yields a duplicate itable containing only the given columns of itbl.
        itbl[rows, cols] is equivalent to itbl[rows][cols] (in fact, rows and cols may be given in
          any order).
        '''
        if cols is not Ellipsis: return self[rows][cols]
        if is_int(rows):
            return self.rows[rows]
        elif isinstance(rows, six.string_types):
            return self.data[rows]
        elif rows is None or len(rows) == 0:
            return ITable(ps.m(), n=0)
        elif isinstance(rows, slice) or is_int(rows[0]):
            n = len(range(rows.start, rows.stop, rows.step)) if isinstance(rows, slice) else \
                len(rows)
            dat = self.data
            def _make_lambda(dat,k): return lambda:dat[k][rows]
            return ITable(
                lazy_map({k:_make_lambda(dat,k) for k in six.iterkeys(dat)}),
                n=n)
        else:
            rows = set(rows)
            dat = self.data
            return ITable(
                reduce(lambda m,k: m if k in rows else m.remove(k), six.iterkeys(dat), dat),
                n=self.row_count)
    def __repr__(self):
        return 'itable(%s, <%d rows>)' % (self.column_names, self.row_count)
    def __iter__(self):
        return six.iterkeys(self.data)
    def __len__(self):
        return len(self.data)
    def __contains__(self, k):
        return ((0 <= k < self.row_count) if is_int(k)  else
                (k in self.data)          if isinstance(k, six.string_types)   else
                False)
    def iterrows(self):
        '''
        itbl.iterrows() iterates over the rows of the givan itable itbl.
        '''
        return iter(self.rows)
def itable(*args, **kwargs):
    '''
    itable(...) yields a new immutable table object from the given set of arguments. The arguments
      may be any number of maps or itables followed by any number of keyword arguments. All the
      entries from the arguments and keywords are collapsed left-to-right (respecting laziness),
      and the resulting column set is returned as the itable. Arguments and maps may contain
      values that are functions of zero arguments; these are considered lazy values and are not
      evaluated by the itable function.
    '''
    # a couple things we want to check first... does our argument list reduce to just an empty
    # itable or just a single itable?
    if len(args) == 0 and len(kwargs) == 0:
        return ITable({}, n=0)
    elif len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], ITable):
        return args[0]
    # we want to try to convert any arguments we can from datatables into maps
    try:
        import pandas
        args = [{k:a[k].values for k in a.keys()} if isinstance(a, pandas.DataFrame) else a
                for a in args]
    except Exception: pass
    # now we want to merge these together and make them one lazy map
    m0 = lazy_map(merge(args, kwargs))
    # see if we can deduce the row size from a non-lazy argument:
    v = m0
    for mm in (list(args) + [kwargs]):
        if len(mm) == 0: continue
        if not is_lazy_map(mm):
            for k in six.iterkeys(mm):
                try: v = mm[k]
                except Exception: continue
                break
        else:
            for k in mm.iternormal():
                try: v = mm[k]
                except Exception: continue
                break
            for k in (mm.itermemoized() if v is m0 else []):
                try: v = mm[k]
                except Exception: continue
                break
        if v is not m0: break
    return ITable(m0, n=(None if v is m0 else len(v)))
def is_itable(arg):
    '''
    is_itable(x) yields True if x is an ITable object and False otherwise.
    '''
    return isinstance(arg, ITable)



####################################################################################################
# pimms/__init__.py
#
# This source-code file is part of the pimms library.
#
# The pimms library is free software: you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with this program.  If
# not, see <http://www.gnu.org/licenses/>.

'''
The pimms library is an immutable data-structure toolkit. It works primarily by decorators applied
to classes and their members to declare how an immutable data-structure's members are related.
Taken together, these decorators form a DSL-like system for declaring immutable data-structures with
full inheritance support.

An immutable data-structure is simply a class that has been modified by the @immutable decorator.
Inside an immutable class, a few things can be declared normally while others must be declared via
the special immutable syntax.
Things that can be delcared normally in an immutable class:
  * Instance methods of the class (def some_method(self, arg): ...)
  * Static methods of the class (using @staticmethod)
  * Static members of the class (some_class_member = 10)
Things that cannot be declared normally in an immutable class:
  * All member variables of a class instance (usually assigned in __init__)
  * __new__, __setattr__, __getattribute__, __delattr__, __copy__, and __deepcopy__ cannot be
    overloaded in immutable classes; doing so will result in undefined behavior
  * Immutable classes should generally only inherit from object or other immutable classes;
    inheritance with other non-immutable classes is fine in theory, especially if only methods are
    added to the class, but member access from immutable objects to non-immutable class members is
    beyond the scope of this library.

Immutable instance member variables, which are usually simply assigned in the class's __init__
method, must be declared specially in immutable classes. All immutable instance members fall into
one of two categories: parameters and values. Parameters are values that must be assigned by the
end of the object's __init__ function, in order for the object to be valid (an exception is raised
if these are not filled). Options are a special kind of parameter that also have default values in
case no assignment is given. Values, unlike parameters, can never be assigned directly; instead,
they are lazily and automatically calculated by a user-provided function of zero or more other
members of the class. Values may depend on either parameters or other values as long as there is
not a circular dependency graph implicit in the declarations.

All such instance member declarations are made using the @param, @option, @value, and @require
decorators, documented briefly here. In all four cases, @param, @option(<default value>), @value,
and @require, the decorator should precede a static function definition.
  * @param declares two things: first, that the name of the static function that follows it is an
    instance member and required parameter of the class and, second, that the static function
    itself, which must take exactly one argument, should be considered a translation function on
    any value assigned to the object; the return value of the function is the value actually
    assigned to the object before any checks are run.
  * @option(<default value>) is identical to @param except that it declares that the given default
    value should be used if no value is assigned to the object in the __init__ method. This value,
    if it is used, is not passed through the translation function that follows.
  * @value declares three things: first, that the name of the static function that follows it is an
    instance member and (lazy) value of the class; second, that the arguments to that static
    function, which must be named exactly after other instance members of the class, are instance on
    members on which this value depends (thus this value will be reset when those members change);
    and third, that the return value of that static function, when given the appropriate member
    values, should be the value assigned to the instance member when requested.
  * @require declares three things: first, that the name of the following static function is the
    identifier for a particular requirement check on the instance members of any object of this
    class; second, that the parameters of that static function, which must match exactly the names
    of other instance members of the class, are the instance members that this requirement checks;
    and third, that the static function's return value will be True if and only if the check is
    passed. The requirement function may throw its own exception or return False, in which case a
    generic exception is raised.

All four decorator types may be overloaded in immutable child classes. Overloading works much as it
does with normal methods; only the youngest child-class's method is required. This can be used to
overload requirements, but new requirements can be placed to act as additional constraints; i.e.,
the youngest class's requirement is always run for all requirements in an object's entire class
hierarchy when a relevant instance members is updated. Overloading may also be used to change an
instance member's type in the child class, such as from a value to a parameter or vice versa. The
child class must, of course, be free of circular dependencies despite these rearrangements.

Note that a required parameter may be implied by the other instance member declarations; this is not
an error and instead inserts the parameter into the class automatically with no translation
function. This occurs when either a value or a requirement declares the parameter as an argument
but no instance member with the parameter's name appears elsewhere in the class or its ancestor
classes. This can be used to create a sort of abstract immutable base class, in that a class may
declare a requirement of some parameter that is not otherwise defined; in order to instantiate the
class, that parameter must be given, either in a child class's __init__ method or as a value or an
explicit (required or optional) parameter.

When an immutable object is constructed, it begins its life in a special 'init' state; this state is
unique in that no requirement checks are run until either the __init__ method returns or a value is
requested of the object; at that point, all non-optional parameters must be specified or an
exception is raised. If all parameters were set, then all requirement checks are run and the
object's state shifts to 'transient'. An immutable object imm can be tested for transience by using
the method imm.is_transient(). A transient object allows its parameters (but never its values) to be
set using normal setattr (imm.x = y) syntax. Requirement checks that are related to a parameter are
changed every time that parameter is set, after the translation function for that parameter is used.
An immutable object remains in the transient state until it is persisted via the imm.persist()
method. Once an object is persistent, it cannot be updated via setattr mechanisms and should be
considered permanent. A new transient duplicate of the object may be created using the
imm.transient() method (this may also be used while the object is still transient). To update the
values of an immutable object, the imm.copy(param1=val1, param2=val2, ...) method should be used.
This method returns a persistent duplicate of the immutable object imm with the given parameters
updated to the given values; these values are always passed through the translation functions
and all relevant chacks are run prior to the return of the copy function. The copy function may
be called on transient or persistent immutables, but the return value is always peresistent.

The additional utility functions are provided as part of the pimms package:
  * is_imm(x) yields True if x is an object that is an instance of an immutable class and False
    otherwise.
  * is_imm_type(x) yields True if x is a class that is immutable and False otherwise.
  * imm_copy(imm, ...) is identical to imm.copy(...) for an immutable object imm.
  * imm_persist(imm) is identical to imm.persist() for a transient immutable object imm.
  * imm_transient(imm) is identical to imm.transient() for an immutable object imm.
  * imm_values(imm_t) yields a list of the values of the immutable class imm_t.
  * imm_params(imm_t) yields a list of the parameters of the immutable class imm_t.
  * imm_dict(imm) is identical to imm.asdict() for an immutable object imm.
  * imm_is_persistent(imm) is identical to imm.is_persistent() for an immutable object imm.
  * imm_is_transient(imm) is identical to imm.is_transient() for an immutable object imm.
'''

from .util        import (lazy_map, LazyPMap, merge, rmerge, lmerge, lmap, is_lmap,
                          lazy_value_map, value_map, key_map, collect, flatten_maps,
                          is_persistent, is_map, is_pmap, is_lazy_map, is_str, is_class,
                          is_nparray, is_npscalar, is_npvector, is_npmatrix, is_npvalue,
                          is_array, is_scalar, is_vector, is_matrix,
                          is_tuple, is_list, is_set, is_seq, is_pseq,
                          is_int, is_float, is_real, is_complex, is_number,
                          is_quantity, is_unit, quant, iquant, unit, mag, like_units, units,
                          imm_array, qhashform, qhash, save, load, io_formats, persist,
                          assoc, dissoc, curry, cache_filename, cache_lmap, cache_fn)
from .immutable   import (immutable, require, value, param, option, is_imm, is_imm_type, imm_copy,
                          imm_persist, imm_transient, imm_params, imm_values, imm_dict,
                          imm_is_persistent, imm_is_transient)
from .calculation import (calc,    plan,    imap,
                          Calc,    Plan,    IMap,
                          is_calc, is_plan, is_imap, planfn)
from .table       import (itable, is_itable, ITable)
from .cmdline     import (argv_parse, argv_parser, to_argv_schema, CommandLineParser,
                          WorkLog, worklog)


def reload_pimms():
    '''
    reload_pimms() reloads the entire pimms module and returns it.
    '''
    import sys, six
    try:              from importlib import reload
    except Exception: from imp       import reload
    reload(sys.modules['pimms.util'])
    reload(sys.modules['pimms.table'])
    reload(sys.modules['pimms.immutable'])
    reload(sys.modules['pimms.calculation'])
    reload(sys.modules['pimms.cmdline'])
    reload(sys.modules['pimms'])
    return sys.modules['pimms']

__version__ = '0.3.10'
description = 'Lazy immutable library for Python built on top of pyrsistent'
